library("RMVL")
#
# In this example we use an input file "points_of_interest.csv" which contains coordinates for several points of interest (f0, ra, dec)
# We use previously constructed spatial index to find points in the O3a atlas that are closest to points of interest
# Results are printed in the end of the file
#

# O3a_atlas.mvl contains results of O3a analysis
Matlas<-mvl_open("O3a_2_atlas.mvl")

# O3a_spatial_index.mvl is created by script create_spatial_index.R from O3a_atlas.mvl
Mspatial<-mvl_open("O3a_2_spatial_index.mvl")

skymaps_band_start<-Mspatial$skymaps_band_start
x0<-Mspatial$x0
x1<-Mspatial$x1
x2<-Mspatial$x2

# User provided files. Example included.
# For very large files, it might be better to store data in MVL format
#points_of_interest<-read.table("points_of_interest.csv", header=TRUE)
points_of_interest<-read.table("hardware_injections.csv", header=TRUE)

atlas_epoch<-1246070000
f0_start_epoch<-1238166018
f0_end_epoch<-1269363618

# Compute frequency at atlas reference epoch with linear approximation
points_of_interest[,"frequency"]<-points_of_interest[,"f0_start"]+(atlas_epoch-f0_start_epoch)*
	(points_of_interest[,"f0_end"]-points_of_interest[,"f0_start"])/(f0_end_epoch-f0_start_epoch)

f0poi<-points_of_interest[,"frequency"]
x0poi<-cos(points_of_interest[,"ra"])*cos(points_of_interest[,"dec"])*f0poi/7.5
x1poi<-sin(points_of_interest[,"ra"])*cos(points_of_interest[,"dec"])*f0poi/7.5
x2poi<-sin(points_of_interest[,"dec"])*f0poi/7.5

compute_polarization_coeffs<-function(iota, psi) {
	ciota<-cos(iota)
	ac<-ciota^2
	ap<-(1+ac)^2/4
	cpsi<-cos(4*psi)
	spsi<-sin(4*psi)
	asum<-0.25*(ap+ac);
	adiff<-0.25*(ap-ac);

	df<-data.frame(pp=asum+adiff*cpsi, pc=2*adiff*spsi, cc=asum-adiff*cpsi, im_pc=0.25*(1+ac)*ciota)

	df[,"pp2"]<-df[,"pp"]^2
	df[,"cc2"]<-df[,"cc"]^2
	df[,"pc2"]<-df[,"pc"]^2
	df[,"ipp"]<-df[,"im_pc"]*df[,"pp"]
	df[,"ipc"]<-df[,"im_pc"]*df[,"pc"]
	df[,"icc"]<-df[,"im_pc"]*df[,"cc"]
	df[,"ppc"]<-df[,"pp"]*df[,"pc"]
	df[,"pcc"]<-df[,"cc"]*df[,"pc"]
	df[,"ppcc"]<-df[,"pp"]*df[,"cc"]
	return(df)
	}

compute_ul<-function(iota, psi, skymap) {
	df<-compute_polarization_coeffs(iota, psi)
	ul2<-skymap[,"unit"]
	for(col in names(df)) {
		ul2<-ul2+df[,col]*skymap[,col]
		}
	return(sqrt(ul2/(df[,"pp"]+df[,"cc"]))*1e-20)
	}


f_closest<-function(i, idx) {
	f0<-f0poi[i]
	# Find frequency bands that cover our point of interest
	fstart<-skymaps_band_start[idx]
	Ffreq<-fstart<=f0 & (fstart+0.045)>=f0
	idx0<-idx[Ffreq]
	if(length(idx0)<1)return(NA)
	
	# Find closest (on the sky) point to our point of interest
	d<-(x0[idx0]-x0poi[i])^2+(x1[idx0]-x1poi[i])^2+(x2[idx0]-x2poi[i])^2
	return(idx0[which.min(d)])
	}

closest_indices<-unlist(mvl_index_lapply(Mspatial$spatial_index, list(f0poi-0.045*0.5, x0poi, x1poi, x2poi), f_closest))

Ffound<-!is.na(closest_indices)
i0<-1:length(f0poi)
ipoi<-i0[Ffound]
iatlas<-closest_indices[Ffound]

df<-data.frame(ipoi=ipoi, iatlas=iatlas, points_of_interest[ipoi,,drop=FALSE], Matlas$skymaps[iatlas, c("ul", "ul_circ", "snr", "snr_frequency", "ra", "dec", "idx"),recurse=TRUE], band_start=skymaps_band_start[iatlas])
names(df)<-gsub(".1", "_falcon", names(df))

# Compute polarization specific upper limits for iota and psi provided in points_of_interest file:
for(i in 1:(dim(df)[1])) {
	df[i,"ul_spec"]<-compute_ul(iota=points_of_interest[df[i, "ipoi"], "iota"],
				psi=points_of_interest[df[i, "ipoi"], "psi"],
				skymap=Matlas$skymaps[df[i, "iatlas"],,drop=FALSE])
	}

#print(df)

# You can write data to a text file with the following command:
# write.table(df, "spatial_index_example2_output.csv", col.names=TRUE, row.names=FALSE, sep="\t")
#

# For hardware injections we can compare recovered values with injection parameters:

df[,"ul_overestimate_pct"]<-100*(df[,"ul_spec"]/df[,"h0"]-1)
df[,"frequency_offset_mHz"]<- (df[,"snr_frequency"]-df[,"frequency"])*1000
df[,"in_paramater_space"]<-abs(df[,"fdot"]) <= 5e-10 & !df[,"binary"]

print(df[,c("pulsar", "frequency", "fdot", "binary", "snr", "ul_overestimate_pct", "frequency_offset_mHz", "in_paramater_space")])

# You can write data to a text file with the following command:
# write.table(df, "spatial_index_example2_hw_inj_comparison.csv", col.names=TRUE, row.names=FALSE, sep="\t")
#




