library("lattice")
library("RMVL")

Matlas<-NULL
try({
	Matlas<-mvl_open("O3a_2_atlas.mvl")
	})
#Msummary<-mvl_open("O3a_2_atlas_summary.mvl")
Msummary<-mvl_open("O3a_2_atlas_summary.mvl")
grid<-mvl2R(Msummary$grid)
Lines<-Msummary$lines[,]
full_spectrum<-mvl2R(Msummary$spectrum)

MAX_X<-max(grid[,"x"])
MAX_Y<-max(grid[,"y"])

outliers<-read.table("O3_outliers.csv", header=TRUE)
outliers[,"RAJD"]<-outliers[,"ra"]*180/pi
outliers[,"DECJD"]<-outliers[,"dec"]*180/pi

full_spectrum[,"outlier"]<-FALSE
for(i in 1:(dim(outliers)[1])) {
	full_spectrum[,"outlier"]<-full_spectrum[,"outlier"] | ((full_spectrum[,"band_start"]<= outliers[i,"frequency"]) & (full_spectrum[,"band_start"]+0.045>=outliers[i,"frequency"]))
	}

injections<-read.table("hardware_injections.csv", header=TRUE)
injections[,"RAJD"]<-injections[,"ra"]*180/pi
injections[,"DECJD"]<-injections[,"dec"]*180/pi

atlas_epoch<-1246070000
f0_start_epoch<-1238166018
f0_end_epoch<-1269363618

# Compute frequency at atlas reference epoch with linear approximation
injections[,"frequency"]<-injections[,"f0_start"]+(atlas_epoch-f0_start_epoch)*
	(injections[,"f0_end"]-injections[,"f0_start"])/(f0_end_epoch-f0_start_epoch)

full_spectrum[,"injection"]<-FALSE
for(i in 1:(dim(injections)[1])) {
	full_spectrum[,"injection"]<-full_spectrum[,"injection"] | ((full_spectrum[,"band_start"]<= injections[i,"frequency"]) & (full_spectrum[,"band_start"]+0.045>=injections[i,"frequency"]))
	}

view_help<-function() {

	cat("\n\tskymap(fstart, fstop, metric)\n")
	cat("\tplot_gw(fstart, fstop, metric, mark.rad, mark.deg, mark.col)\n")
	cat("\n")
	cat("The function skymap() returns a data.frame with sky map data, while plot_gw() plots it.\n")
	cat("Parameters:\n")
	cat("\tfstart      - frequency band start\n")
	cat("\tfstop       - frequency band stop\n")
	cat("\tmetric      - either \"snr\" (for SNR) or \"ul\" (for worst-case upper limit)\n")
	cat("\tmark.rad    - coordinates of optional mark to add to the plot c(ra=RA, dec=Dec), in radians, J2000\n")
	cat("\tmark.deg    - coordinates of optional mark to add to the plot c(ra=RA, dec=Dec), in degrees, J2000\n")
	cat("\tmark.col    - color of optional mark, defaults to \"red\"\n")

	cat("\nExamples:\n\n")
	cat("\tplot_gw(835, 840, \"snr\")\n\n")
	cat("\tplot_gw(835, 840, \"ul\", mark.deg=c(ra=45, dec=0), mark.col=\"magenta\")\n\n")
	cat("\tdf<-skymap(835, 840, \"snr\")\n\n")
	}

Lines<-Msummary$lines[,]

epsilon<-23.439281*pi/180
# x, y, z
ecliptic_pole<-c(0, -sin(epsilon), cos(epsilon))

galactic_ra<-(12+51.4/60)*(2*pi)/24
galactic_dec<-27.17*pi/180
galactic_pole<-c(cos(galactic_dec)*cos(galactic_ra), cos(galactic_dec)*sin(galactic_ra), sin(galactic_dec))

plot_x<-function(ra, dec) { return(180+(180-ra)*cos(dec*pi/180.0)) }
limit_y<-function(y) { return(ifelse(abs(y)>90, NA, y)) }

plot_band<-function(v, width, ...) {
	v<-v/sqrt(sum(v*v))
	ra0<-seq(0, 360, by=1)

	x1<-cos(ra0*pi/180)*v[1]
	y1<-sin(ra0*pi/180)*v[2]
	a1<-x1+y1
	z1<-v[3]

	w1a<- cos(pi/2-(0.5*width/180)*pi)
	w1b<- cos(pi/2+(0.5*width/180)*pi)

	n1<-sqrt(a1^2+z1^2)

	a2<-a1/n1
	z2<-z1/n1
	w2a<-w1a/n1
	w2b<-w1b/n1

	angle3<-atan2(z2, a2)*180/pi
	dec0a<- acos(w2a)*180/pi+angle3
	dec0b<- acos(w2b)*180/pi+angle3

	points(plot_x(ra0, dec0a)/360, 0.5+limit_y(dec0a)/180, type="l", ...)
	points(plot_x(ra0, dec0b)/360, 0.5+limit_y(dec0b)/180, type="l", ...)

	dec0a<- -acos(w2a)*180/pi+angle3
	dec0b<- -acos(w2b)*180/pi+angle3

	points(plot_x(ra0, dec0a)/360, 0.5+limit_y(dec0a)/180, type="l", ...)
	points(plot_x(ra0, dec0b)/360, 0.5+limit_y(dec0b)/180, type="l", ...)
	}

plot_points<-function(RAJD, DECJD, ...) {
	points(plot_x(RAJD, DECJD)/360, 0.5+limit_y(DECJD)/180, ...)
	}

plot_ecliptic_band<-function(width, ...) {
	plot_band(ecliptic_pole, width, ...)
	}

plot_galactic_band<-function(width, ...) {
	plot_band(galactic_pole, width, ...)
	}

compute_polarization_coeffs<-function(iota, psi) {
	ciota<-cos(iota)
	ac<-ciota^2
	ap<-(1+ac)^2/4
	cpsi<-cos(4*psi)
	spsi<-sin(4*psi)
	asum<-0.25*(ap+ac);
	adiff<-0.25*(ap-ac);

	df<-data.frame(pp=asum+adiff*cpsi, pc=2*adiff*spsi, cc=asum-adiff*cpsi, im_pc=0.25*(1+ac)*ciota)

	df[,"pp2"]<-df[,"pp"]^2
	df[,"cc2"]<-df[,"cc"]^2
	df[,"pc2"]<-df[,"pc"]^2
	df[,"ipp"]<-df[,"im_pc"]*df[,"pp"]
	df[,"ipc"]<-df[,"im_pc"]*df[,"pc"]
	df[,"icc"]<-df[,"im_pc"]*df[,"cc"]
	df[,"ppc"]<-df[,"pp"]*df[,"pc"]
	df[,"pcc"]<-df[,"cc"]*df[,"pc"]
	df[,"ppcc"]<-df[,"pp"]*df[,"cc"]
	return(df)
	}

compute_ul<-function(iota, psi, skymap) {
	df<-compute_polarization_coeffs(iota, psi)
	ul2<-skymap[,"unit"]
	for(col in names(df)) {
		ul2<-ul2+df[,col]*skymap[,col]
		}
	return(sqrt(ul2/(df[,"pp"]+df[,"cc"]))*1e-20)
	}


atlas_max<-function(fstart, fstop, center=NULL, width=0.1, col="snr", idx=NULL, iota1="snr", psi1="snr") {
	if(is.null(Matlas))return(NULL)

	params<-Matlas$parameters

	if(!is.null(center)) {
		fstart<-center-width
		fstop<-center+width
		}

	if(is.null(idx)) {
		idx<-params[params[,"band_start"][]+0.045>=fstart & (params[,"band_start"][]<fstop), "idx"]
		}

	cv<-Matlas$skymaps[,col,ref=TRUE]

	idx2<-unlist(mvl_index_lapply(Matlas["skymaps_extent_index_idx"], list(idx), function(i, idx) {
		v<-cv[idx]
		return(idx[v==max(v)])
		}))
	idx3<-max(Matlas$skymaps[idx2, col])==Matlas$skymaps[idx2, col]
	sk<-Matlas$skymaps[idx2[idx3],,drop=FALSE]
	sk[,"RAJD"]<-sk[,"ra"]*180/pi
	sk[,"DECD"]<-sk[,"dec"]*180/pi

	if(iota1=="snr") {
		iota1<-sk[,"snr_iota"]
		}

	if(psi1=="snr") {
		psi1<-sk[,"snr_psi"]
		}

	sk[,"ul_spec"]<-compute_ul(iota1, psi1, sk)
	sk[,"resolution"]<-Matlas$parameters[sk[,"idx"], "resolution"]
	sk[,"resolutionM"]<-sk[,"resolution"]*180*60/pi

	return(sk)
	}

# band_image<-function(first_bin) {
# 	M<-matrix(NA, max(mvl2R(Msummary$grid[,"x"])), max(mvl2R(Msummary$grid[,"y"])))
# 	data<-Msummary$summary_skymaps[mvl2R(Msummary$summary_skymaps[,"first_bin"])==first_bin,]
# 	for(i in 1:(dim(data)[1])) M[grid[data$grid_idx[i], "x"], grid[data$grid_idx[i], "y"]]<-data[i,"snr"]
# 	return(M)
# 	}

grid_x<-grid[,"x"]
grid_y<-grid[,"y"]

skymap_to_image<-function(sk, col) {
	display_idx<-grid_x[sk$grid_idx]+(grid_y[sk$grid_idx]-1)*MAX_X
	M<-array(NA, MAX_X*MAX_Y)
	M[display_idx]<-sk[,col]
	dim(M)<-c(MAX_X, MAX_Y)
	return(M)
	}

get_band<-function(first_bin, col) {
	idx<-mvl_index_lapply(Msummary$summary_skymaps_extent_index_first_bin, list(first_bin), function(i, idx) { return(idx) } )[[1]]
	if(is.null(idx))return(NULL)
	data<-mvl2R(Msummary$summary_skymaps[idx, c("grid_idx", col),drop=FALSE])
	return(data)
	}

band_image<-function(first_bin, col) {
#	data<-Msummary$summary_skymaps[mvl2R(Msummary$summary_skymaps[,"first_bin"])==first_bin,]
	idx<-mvl_index_lapply(Msummary$summary_skymaps_extent_index_first_bin, list(first_bin), function(i, idx) { return(idx) } )[[1]]
	if(is.null(idx)) {
		M<-array(NA, MAX_X*MAX_Y)
		#M[display_idx]<-data[,col]
		dim(M)<-c(MAX_X, MAX_Y)
		return(M)
		}
#	cat("length(idx)=", length(idx), "\n")
	data<-Msummary$summary_skymaps[idx, c("grid_idx", col)]

	return(skymap_to_image(data, col))
#
# 	display_idx<-grid_x[data$grid_idx]+(grid_y[data$grid_idx]-1)*MAX_X
# 	M<-array(NA, MAX_X*MAX_Y)
# 	M[display_idx]<-data[,col]
# 	dim(M)<-c(MAX_X, MAX_Y)
# 	return(M)
	}

pmax_skymap<-function(df1, df2, col) {
	if(is.null(df1))return(df2)
	if(is.null(df2))return(df1)

	if(any(df1[,"grid_idx"]!=df2[,"grid_idx"]))stop("skymap mismatch")
	df1[,col]<-pmax(df1[,col], df2[,col], na.rm=TRUE)
	return(df1)
	}

skymap<-function(fstart=0, fend=1e5, col) {
	fbins<-full_spectrum[(full_spectrum[,"band_start"]<=fend) & (full_spectrum[,"band_stop"]>=fstart), "first_bin"]
	if(length(fbins)<1)return(NULL)

	cat("Using:\n")
	df<-NULL
	for(n in c("summary_100Hz", "summary_10Hz", "summary_1Hz")) {
		cat("\t", n, "\n")
		L<-Msummary[n]
		cnt<-L[["contents"]]
		F<-(cnt[,"min_fbin"] %in% fbins) & (cnt[,"max_fbin"] %in% fbins)

		for(k in (1:(dim(cnt)[1]))[F]) {
			fb<-mvl2R(L[["fbins"]][k][[1]])
			if(any(! fb %in% fbins)) {
				browser()
				stop("Missing fbins")
				}
			df<-pmax_skymap(df, mvl2R(L[["data"]][k][[1]]), col)
			fbins<-setdiff(fbins, fb)
			}
		}

	cat("\tsingles\n")
	if(length(fbins)>0)
	for(i in 1:length(fbins)) {
		#	if(i %% 100==0)cat(".")
			#M<-pmax(M, band_image(fbins[[i]], col), na.rm=TRUE)
			df<-pmax_skymap(df, get_band(fbins[[i]], col), col)
			}
#	browser()
	return(skymap_to_image(df, col))
	}


# formatting function
# https://stackoverflow.com/questions/10762287/how-can-i-format-axis-labels-with-exponents-with-ggplot2-and-scales
scientific <- function(x) {
    if(any(regexpr("e", x)>=0))return(ifelse(x==0, "0", parse(text=gsub("[+]", "", gsub("e", " %*% 10^", x)))))
	else return(x)
}

color.bar<-function(lut, min, max=-min, nticks=11, ticks=NULL, title=NA, ...) {
	scale <- (length(lut)-1)/(max-min)

	if(is.null(ticks)) {
		ticks<-pretty(c(min, max), n=nticks)
		ticks<-c(min, ticks[ticks>min & ticks<max], max)

		ticks<-seq(min, max, length=nticks)
		}

	plot(c(0,10), c(min,max), type='n', bty='n', xaxt='n', xlab=NA, yaxt='n', ylab=NA, main=title)
	axis(2, at=ticks, labels=scientific(format(ticks)), las=1, ...)
	for(i in 1:(length(lut)-1)) {
		y<-(i-1)/scale + min
		rect(0,y,10,y+1/scale, col=lut[i], border=NA)
		}
	}

custom_legend<-function(title=NA, ...) {
	plot(c(0,100), c(0,100), type='n', bty='n', xaxt='n', xlab=NA, yaxt='n', ylab=NA, main=title)
	legend(0, 100, c("outlier", "injection", "line", "ecl_max"), col=c("blue", "black", "red", "magenta"), pch=c(5, 6, 8, 1))
	}

plot_skymap<-function(sk, col, levels=NULL, ...) {
# 	N_x<-max(sk[,"x"])
# 	N_y<-max(sk[,"y"])
# 	M<-matrix(NA, N_x, N_y)
#  	for(i in 1:(dim(sk)[1]))
#  		M[sk[i,"x"], sk[i,"y"]]<-sk[i,col]

	cols<-rev(rainbow(50, end=0.75))
	image(sk, col=cols, xaxt= "n", yaxt= "n", xlab="RA", ylab="Dec", ...)

	axis(1, at=seq(0, 1, 1/12), labels=rev(seq(0, 360, by=30)), ...)
	axis(2, at=seq(0, 1, 1/6), labels=(seq(-90, 90, by=30)), ...)

	}

rational_levels<-function(x) {
	levels<-quantile(x, c(0, 1), na.rm=TRUE)
	M<-max(abs(levels-mean(levels)))
	if(M>0) {
		M1<-10^(ceiling(log10(2*M))-2)
		levels[1]<-floor(levels[1]/M1)*M1
		levels[2]<-ceiling(levels[2]/M1)*M1
		}
	return(levels)
	}


plot_gw<-function(fstart=0, fend=1e5, col="snr", center=NULL, width=0.1, galactic.band=TRUE, plot.outliers=TRUE, plot.injections=TRUE, mark.rad=NULL, mark.deg=NULL, mark.col="red", n_spectrum_peaks=10, ...) {
	if(! col %in% c("snr", "ul")) stop("Parameter col=\"", col, "\" must be either \"snr\" or \"ul\"")

	if(!is.null(center)) {
		fstart<-center-width
		fend<-center+width
		}

	if(fstart>fend)stop("Invalid frequency range, fstart>fend")

	sk<-skymap(fstart=fstart, fend=fend, col=col)
	layout(matrix(c(1,2,3,4), 2, 2, byrow = TRUE), width=c(17, 3), heights=c(2,1))
	plot_skymap(sk, col, main=ifelse(col=="snr", expression("SNR"), expression(paste(h[0], " upper limit"))), ...)

	if(galactic.band)
		plot_galactic_band(25, col="red", lwd=2)

	if(plot.outliers) {
		Fp<-outliers[,"snr"]>0
		Fp[is.na(Fp)]<-TRUE
		Fp<-Fp & outliers[,"frequency"]>=fstart & outliers[,"frequency"]<=fend+0.045
		Fp[is.na(Fp)]<-FALSE
		if(any(Fp)) {
			cat(sum(Fp), "outliers in band\n")
			Pdf<-outliers[Fp,,drop=FALSE]
			if(dim(Pdf)[1]<10)print(Pdf)
			plot_points(Pdf[,"RAJD"], Pdf[,"DECJD"], pch=5, col="blue", cex=2)
			}
		}

	if(plot.injections) {
		Fp<-injections[,"frequency"]>=fstart & injections[,"frequency"]<=fend+0.045
		Fp[is.na(Fp)]<-FALSE
		if(any(Fp)) {
			cat(sum(Fp), "injections in band\n")
			Pdf<-injections[Fp,,drop=FALSE]
			if(dim(Pdf)[1]<10)print(Pdf)
			plot_points(Pdf[,"RAJD"], Pdf[,"DECJD"], pch=6, col="black", cex=2)
			}
		}

	if(!is.null(mark.deg)) {
		plot_points(mark.deg[["ra"]], mark.deg[["dec"]], pch=5, col=mark.col, cex=2)
		}

	if(!is.null(mark.rad)) {
		plot_points(mark.rad[["ra"]]*180/pi, mark.rad[["dec"]]*180/pi, pch=5, col=mark.col, cex=2)
		}

	pp<-par(mar=c(2, 4, 2, 1))


	levels<-rational_levels(sk)

	color.bar(rev(rainbow(50, end=0.75)), levels[[1]], levels[[2]], ...)
	par(pp)

	pp<-par(mar=c(5, 4, 1, 2)+0.1)
	sp<-full_spectrum
	Fband<-sp[,"band_stop"]>=fstart & sp[,"band_start"]<=fend
	sp2<-sp[Fband,,drop=FALSE]
	bnd<-rational_levels(sp[Fband, col])
	if(sum(Fband)>20) {
		plot(sp2[,"band_start"], sp2[, col], type="l", xlab="Frequency, Hz", ylab=col, yaxt="n", col="black", ylim=bnd,  ...)
		X<-sp2[, col]
		X[sp2[, "line_mark"]<1]<- NA
		points(sp2[,"band_start"], X, type="o", xlab="Frequency, Hz", ylab=col, pch=8, col="red", ...)

		X<-sp2[, col]
		X[sp2[, "snr"]>sp2[, "snr_ecl"]]<- NA
		points(sp2[,"band_start"], X, type="o", xlab="Frequency, Hz", ylab=col, col="magenta", ...)

		}
		else {
		plot(sp2[,"band_start"], sp2[, col], type="o", xlab="Frequency, Hz", ylab=col, yaxt="n", col="black", ylim=bnd,  ...)

		X<-sp2[, col]
		X[sp2[, "line_mark"]<1]<- NA
		points(sp2[,"band_start"], X, type="o", xlab="Frequency, Hz", ylab=col, pch=8, col="red", ...)

		X<-sp2[, col]
		X[sp2[, "snr"]>sp2[, "snr_ecl"]]<- NA
		points(sp2[,"band_start"], X, type="o", xlab="Frequency, Hz", ylab=col, col="magenta", ...)
		}

	ticks<-seq(bnd[[1]], bnd[[2]], length=5)
	axis(2, at=ticks, labels=ifelse(c(TRUE, rep(FALSE, length(ticks)-2), TRUE),scientific(format(ticks)), NA))

	if(plot.outliers) {
		X<-sp[Fband, col]
		FF<-sp[Fband, "band_start"]
		Fn0<-outliers[,"snr"]>0
		Fn0[is.na(Fn0)]<-TRUE
		if(length(X)>0) {
			X[!sp[Fband, "outlier"]]<-NA
# 			for(i in 1:(length(X))) {
# 				Fn<-Fn0 & (FF[i]<= outliers[,"frequency"]) & (FF[i]+0.045>=outliers[,"frequency"])
# 				Fn[is.na(Fn)]<-FALSE
# 				if(!any(Fn))X[i]<-NA
# 				}
			}
		points(sp[Fband,"band_start"], X, type="o", pch=5, col="blue", ...)
		}
	if(plot.injections) {
		X<-sp[Fband, col]
		FF<-sp[Fband, "band_start"]
		out_of_range<-FALSE
		if(length(X)>0) {
			X[!sp[Fband, "injection"]]<-NA
			for(i in (1:(length(X)))[!is.na(X)]) {
				Fn<-(FF[i]<= injections[,"frequency"]) & (FF[i]+0.045>=injections[,"frequency"])
				Fn[is.na(Fn)]<-FALSE
#				if(!any(Fn))X[i]<-NA
				if(any(abs(injections[Fn, "fdot"])>5e-10 | injections[Fn, "binary"]))out_of_range<-TRUE
				}
			}
		points(sp[Fband,"band_start"], X, type="o", pch=6, col="black", ...)
		if(out_of_range)
			cat("Some injections are outside searched parameter space\n")
		}
	par(pp)

	pp<-par(mar=c(0, 0, 1, 1), mgp=c(0, 0, 0))
	custom_legend()
	par(pp)

	cat("Spectrum peaks:\n")
	print(sp2[order(sp2[,col], decreasing=TRUE),,drop=FALSE][1:min(n_spectrum_peaks, dim(sp2)[1]),,drop=FALSE])

	if(abs(fend-fstart)< 1) {
		sk2<-atlas_max(fstart=fstart, fstop=fend, col=col)
		if(is.null(sk2)) {
			cat("O3a_2_atlas.mvl is not loaded, cannot display atlas maximum\n")
			} else {
			cat("Maximum of" ,col, "is achieved at:\n")
			sk2[,"RAJD"]<-180*sk2[,"ra"]/pi
			sk2[,"DECD"]<-180*sk2[,"dec"]/pi
			print(sk2[,c("idx", "ra", "dec", "resolution", "RAJD", "DECD", "resolutionM", "ul", "ul_spec", "snr", "snr_frequency", "snr_iota", "snr_psi"),drop=FALSE])
			}
		}
	}


epsilon<-23.439281*pi/180
# x, y, z
ecliptic_pole<-c(0, -sin(epsilon), cos(epsilon))

galactic_ra<-(12+51.4/60)*(2*pi)/24
galactic_dec<-27.17*pi/180
galactic_pole<-c(cos(galactic_dec)*cos(galactic_ra), cos(galactic_dec)*sin(galactic_ra), sin(galactic_dec))

plot_x<-function(ra, dec) { return(180+(180-ra)*cos(dec*pi/180.0)) }
limit_y<-function(y) { return(ifelse(abs(y)>90, NA, y)) }


epsilon<-23.439281*pi/180
# x, y, z
ecliptic_pole<-c(0, -sin(epsilon), cos(epsilon))

galactic_ra<-(12+51.4/60)*(2*pi)/24
galactic_dec<-27.17*pi/180
galactic_pole<-c(cos(galactic_dec)*cos(galactic_ra), cos(galactic_dec)*sin(galactic_ra), sin(galactic_dec))

plot_x<-function(ra, dec) { return(180+(180-ra)*cos(dec*pi/180.0)) }
limit_y<-function(y) { return(ifelse(abs(y)>90, NA, y)) }


# Change to TRUE to produce skymaps from the paper
if(FALSE) {

pdf("skymap_835_840_snr.pdf", width=12, height=10, pointsize=20)
plot_gw(835, 840, "snr")
dev.off()

pdf("skymap_835_840_ul.pdf", width=12, height=10, pointsize=20)
plot_gw(835, 840, "ul")
dev.off()

pdf("skymap_1220_snr.pdf", width=12, height=10, pointsize=20)
plot_gw(1220.2, 1220.4, "snr")
dev.off()

pdf("skymap_1220_ul.pdf", width=12, height=10, pointsize=20)
plot_gw(1220.2, 1220.4, "ul")
dev.off()

pdf("skymap_763_764_ip9_ul.pdf", width=12, height=10, pointsize=20)
plot_gw(763, 764, "ul")
dev.off()

pdf("skymap_763_3_763_42_ul.pdf", width=12, height=10, pointsize=20)
plot_gw(763.3, 763.42, "ul")
dev.off()
}



plot_gw(835, 840, "snr")

cat("\nCall view_help() for examples and usage\n\n")
